#pragma link "ch375dll.lib" //link against the import lib
#include "ch375dll.h"

#include <stdio.h>
#include <string.h>
#include <conio.h>

/* firmware buffer */
unsigned char fbuffer[0x4000];

/* table for speeding up fw reads*/
const unsigned char stats_table[256] =
{
	 0xff, 0x02, 0x00, 0xf5, 0xe5, 0x75, 0x03, 0x04, 
	 0x80, 0x05, 0xd2, 0x01, 0xe4, 0xef, 0x82, 0x83, 
   0x08, 0x24, 0xc2, 0x60, 0xe0, 0x12, 0x7f, 0x34, 
   0x10, 0x07, 0x22, 0x40, 0x54, 0x94, 0x30, 0x70, 
   0xc0, 0xf0, 0xaf, 0xd0, 0x44, 0xa3, 0x36, 0x74, 
   0x15, 0xc3, 0x09, 0x93, 0x53, 0xec, 0x48, 0x06, 
   0x0a, 0x14, 0x20, 0x25, 0x50, 0x64, 0xd4, 0x16, 
   0x43, 0x47, 0xd6, 0xe7, 0xea, 0x0c, 0x32, 0x3f, 
   0x46, 0x90, 0xc8, 0xdf, 0x38, 0x45, 0xb4, 0xd3, 
   0xfa, 0xa1, 0xc5, 0xca, 0xcc, 0xde, 0xfc, 0x0b, 
   0x23, 0x37, 0x42, 0xed, 0xfb, 0x2f, 0x95, 0x55, 
   0x85, 0xdc, 0x18, 0x26, 0x33, 0x7d, 0x89, 0xac, 
   0xae, 0xfe, 0x0f, 0x17, 0x1b, 0x27, 0x35, 0x39, 
   0x3e, 0x57, 0x78, 0x8f, 0xa9, 0xaa, 0xc1, 0xd9, 
   0xdd, 0xe3, 0xf3, 0xf8, 0x0d, 0x21, 0x3b, 0x3c, 
   0x73, 0x81, 0x87, 0x88, 0x8a, 0x99, 0xbf, 0xdb, 
   0xf2, 0xfd, 0x1a, 0x1f, 0x31, 0x5f, 0x6c, 0x7a, 
   0x7e, 0x8e, 0xbc, 0xd5, 0xd8, 0xda, 0xe9, 0xeb, 
   0xee, 0xf6, 0x11, 0x1c, 0x29, 0x2d, 0x56, 0x58, 
   0x7c, 0x8d, 0x91, 0x98, 0xb3, 0xb9, 0xd7, 0xe1, 
   0xe6, 0xe8, 0xf9, 0x13, 0x1e, 0x28, 0x2e, 0x41, 
   0x4e, 0x69, 0x79, 0x7b, 0x9e, 0x9f, 0xa0, 0xab, 
   0xad, 0xcf, 0xe2, 0x0e, 0x19, 0x1d, 0x2a, 0x4b, 
   0x52, 0x5b, 0x63, 0x84, 0x86, 0x8c, 0x9d, 0xa2, 
   0xb1, 0xb2, 0xc4, 0x2b, 0x49, 0x4a, 0x4c, 0x4d, 
   0x59, 0x61, 0x67, 0x68, 0x6b, 0x6d, 0x6e, 0x6f, 
   0x77, 0x92, 0x96, 0x9a, 0xa6, 0xa8, 0xb0, 0xb5, 
   0xbb, 0xc6, 0xc7, 0xc9, 0xcd, 0xd1, 0xf4, 0x2c, 
   0x3a, 0x3d, 0x4f, 0x51, 0x5a, 0x5c, 0x5d, 0x5e, 
   0x62, 0x65, 0x66, 0x6a, 0x71, 0x72, 0x76, 0x8b, 
   0x97, 0x9b, 0x9c, 0xa4, 0xa5, 0xa7, 0xb6, 0xb7, 
   0xb8, 0xba, 0xbd, 0xbe, 0xcb, 0xce, 0xf1, 0xf7
};

/* a simple function for showing the program is still alive */
void IsAlive(unsigned short Address)
{
   int static last;
   int current = Address >> 6;
   if (current != last)
   {
      fprintf(stderr, ".");
      last = current;
   }
}

/* try to find a suitable device*/
int SearchDevice(void)
{
   int devid = CH375OpenDevice(-1);
   if (devid !=-1) CH375CloseDevice(devid);
   if (CH375GetUsbID(devid) == 0x55E04348)
     return devid;
   return -1;     
}

/* test the bootloader version */
int CheckLoaderVer(int DevId)
{
   unsigned int rxlen,txlen;
   char TxBuffer[40];
   char RxBuffer[40];
   int  result;   
   
   TxBuffer[0]= 0xBB;
   txlen=1;
   result = CH375WriteEndP(DevId,2,TxBuffer,&txlen);
   if (result)
   {
      result = CH375ReadEndP(DevId,2,RxBuffer,&rxlen);
      if ((result) && (rxlen))
        return RxBuffer[0];
   }
   return -1;
}

/* change the scrmble key */
int InitKey(int DevId, unsigned int Key)
{
   unsigned int rxlen,txlen;
   char TxBuffer[40];
   char RxBuffer[40];
   int  result;   
   
   TxBuffer[0]=  0xA6;
   TxBuffer[1] = 4;
   TxBuffer[2] = Key>>24 & 0xFF; 
   TxBuffer[3] = Key>>16 & 0xFF;
   TxBuffer[4] = Key>> 8 & 0xFF; 
   TxBuffer[5] = Key>> 0 & 0xFF;
   txlen=0x06;
   result = CH375WriteEndP(DevId,2,TxBuffer,&txlen);
   if (result)
   {
      result = CH375ReadEndP(DevId,2,RxBuffer,&rxlen);
      if ((result) && (rxlen))
        return RxBuffer[0];
   }
   return -1;
}

/* look for 16 0xFF in the code */
int FindFFRecord(int DevId)
{ 
   unsigned int rxlen,txlen;
   char TxBuffer[40];
   char RxBuffer[40];
   int  result;   
   unsigned short address = 0x3FF0;
   
   if (InitKey(DevId,0) == -1)
   {
      fprintf(stderr,"failed to reset key\n");
      return -1; 
   }
   memset(TxBuffer,0xFF,sizeof(TxBuffer));
   TxBuffer[0] = 0xA7;
   TxBuffer[1] = 16;
   TxBuffer[2] = address >> 0 & 0xFF;
   TxBuffer[3] = address >> 8 & 0xFF;
   txlen =0x40;
   while (address)
   {
      result = CH375WriteEndP(DevId,2,TxBuffer,&txlen);
      if (result)
      {
         result = CH375ReadEndP(DevId,2,RxBuffer,&rxlen); 
         if ((result) && (rxlen))
         {
            if (RxBuffer[0]==0) break;
         }
      }
      address--;
      TxBuffer[2] = address >> 0 & 0xFF;
      TxBuffer[3] = address >> 8 & 0xFF;    
   }
   if (address) return address;
   return -1;  
}

/* read out the flash */
int AttackFlash(int DevId,int Address)
{
	 unsigned int rxlen,txlen;
   char TxBuffer[40];
   char RxBuffer[40];
   int  result;  
   int  i; 

	 memset(fbuffer,0xFF,sizeof(fbuffer));

   memset(TxBuffer,0xFF,sizeof(TxBuffer));
   TxBuffer[0] = 0xA7;
   TxBuffer[1] = 16;
   TxBuffer[2] = Address >> 0 & 0xFF;
   TxBuffer[3] = Address >> 8 & 0xFF;
   txlen =0x40;
   while (Address)
   {
   	  for (i=0;i<256;i++)
   	  {
   	  	 TxBuffer[4] = stats_table[i];
         result = CH375WriteEndP(DevId,2,TxBuffer,&txlen);
         if (result)
         {
            result = CH375ReadEndP(DevId,2,RxBuffer,&rxlen); 
            if ((result) && (rxlen))
            {
               if (RxBuffer[0]==0) 
               {
               	  fbuffer[Address]=TxBuffer[4]; 
               	  break;
               }
            } else return -1;
         } else return -1;
      }
      memmove(&TxBuffer[5],&TxBuffer[4],16);
      Address--;
      TxBuffer[2] = Address >> 0 & 0xFF;
      TxBuffer[3] = Address >> 8 & 0xFF;
      if (kbhit())
      {
      	 if (getch()==27) return -1; 
      }	
      IsAlive(Address);       	
   }   
	 return 0;
}

/*  store the contents to a Intel hex file*/
int WriteIntelHex(char *filename)
{
   FILE     *file;
   char     line[80];
   unsigned char  nbytes,b,chk;
   unsigned short addr;
   int      i;

   if (filename)
   {
      file = fopen(filename,"wt");
      if (!file)
      {
         fprintf(stderr, "creating hexfile failed!\n");
         return -1 ;
      }
      addr=0;
      while (addr < 0x3FF0)
      {
      	 line[0]= ':';
      	 line[1]= '1';
      	 line[2]= '0';      	 
      	 chk= 16;
      	 sprintf(&line[3],"%04X",addr); chk+= addr & 0xFF;chk+= addr >>8 & 0xFF;
      	 line[7]='0'; line[8]='0';chk+=0;
      	 for (i=0;i<16;i++)
      	 {
      	 	  sprintf(&line[i*2+9],"%02X",fbuffer[addr+i]);
      	 	  chk+= fbuffer[addr+i];
      	 }
      	 sprintf(&line[i*2+9],"%02X",chk); 
      	 line[i*2+11]='\0';
      	 fprintf (file,"%s\n",line);
      	 addr +=16;
      }
      fprintf (file,":00000001FF\n");
      fclose(file);
      return 0;
   }
   return -1;
}

int WriteBinImage(char *filename)
{
   FILE     *file;
   int      addr;

   if (filename)
   {
      file = fopen(filename,"wb");
      if (!file)
      {
         fprintf(stderr, "creating binary file failed!\n");
         return -1 ;
      }
      addr=0;
      while (addr < 0x3FF0)
      {
      	 fwrite(&fbuffer[addr],1,1,file);
      	 addr++;  
      }
      fclose(file);
   }
   return 0;
}


int main(void) 
{
   int address;
   int id = SearchDevice();
   fprintf(stderr,"CH552 Flash Dumper\n");

   if(id==-1)
   {
      fprintf(stderr,"no device found!\n");
      return -1; 
   }
   
   id =CH375OpenDevice(id); 
   if(CheckLoaderVer(id)!=0x11)
   {
      fprintf(stderr,"wrong loader version... v1.1 needed!\n");
      CH375CloseDevice(id);
      return -1; 
   }  
   
   address=FindFFRecord(id); 
   if (address==-1)
   {
      fprintf(stderr,"no starting point found!\n");
      CH375CloseDevice(id);
      return -1;
   }
   else
   {
      fprintf(stderr,"... starting point found at 0x%4.4X\n",address);
      fprintf(stderr,"... now reading the flash. This may take some time. If bored cancel with ESC\n");
      AttackFlash(id,address);
      WriteIntelHex("firmware.hex");
      WriteBinImage("firmware.bin");
   }  
   CH375CloseDevice(id);  
   return 0;
}
 